##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from typing import Dict, List, Union, Optional, Any, Set, cast
from warnings import warn

import numpy as np

from pyquil.api._base_connection import ForestConnection
from pyquil.api._error_reporting import _record_call
from pyquil.paulis import PauliSum, PauliTerm
from pyquil.quil import Program, percolate_declares
from pyquil.quilatom import MemoryReference
from pyquil.gates import MOVE
from pyquil.wavefunction import Wavefunction


class WavefunctionSimulator:
    @_record_call
    def __init__(
        self, connection: Optional[ForestConnection] = None, random_seed: Optional[int] = None
    ) -> None:
        """
        A simulator that propagates a wavefunction representation of a quantum state.

        :param connection: A connection to the Forest web API.
        :param random_seed: A seed for the simulator's random number generators. Either None (for
            an automatically generated seed) or a non-negative integer.
        """
        if connection is None:
            connection = ForestConnection()

        self.connection = connection

        if random_seed is None:
            self.random_seed = None
        elif isinstance(random_seed, int) and random_seed >= 0:
            self.random_seed = random_seed
        else:
            raise TypeError("random_seed should be None or a non-negative int")

    @_record_call
    def wavefunction(self, quil_program: Program, memory_map: Any = None) -> Wavefunction:
        """
        Simulate a Quil program and return the wavefunction.

        .. note:: If your program contains measurements or noisy gates, this method may not do what
            you want. If the execution of ``quil_program`` is **non-deterministic** then the
            final wavefunction only represents a stochastically generated sample and the
            wavefunctions returned by *different* ``wavefunction`` calls *will generally be
            different*.

        :param quil_program: A Quil program.
        :param memory_map: An assignment of classical registers to values, representing an initial
                           state for the QAM's classical memory.

                           This is expected to be of type Dict[str, List[Union[int, float]]],
                           where the keys are memory region names and the values are arrays of
                           initialization data.

                           For now, we also support input of type Dict[MemoryReference, Any],
                           but this is deprecated and will be removed in a future release.
        :return: A Wavefunction object representing the state of the QVM.
        """

        if memory_map is not None:
            quil_program = self.augment_program_with_memory_values(quil_program, memory_map)

        return cast(
            Wavefunction,
            self.connection._wavefunction(quil_program=quil_program, random_seed=self.random_seed),
        )

    @_record_call
    def expectation(
        self,
        prep_prog: Program,
        pauli_terms: Union[PauliSum, List[PauliTerm]],
        memory_map: Any = None,
    ) -> Union[float, np.ndarray]:
        """
        Calculate the expectation value of Pauli operators given a state prepared by prep_program.

        If ``pauli_terms`` is a ``PauliSum`` then the returned value is a single ``float``,
        otherwise the returned value is an array of values, one for each ``PauliTerm`` in the
        list.

        .. note:: If your program contains measurements or noisy gates, this method may not do what
            you want. If the execution of ``quil_program`` is **non-deterministic** then the
            final wavefunction from which the expectation value is calculated only represents
            a stochastically generated sample and the wavefunctions returned by *different*
            ``wavefunction`` calls *will generally be different*.

        :param prep_prog: A program that prepares the state on which we measure the expectation.
        :param pauli_terms: A Pauli representation of a quantum operator.
        :param memory_map: An assignment of classical registers to values, representing an initial
                           state for the QAM's classical memory.

                           This is expected to be of type Dict[str, List[Union[int, float]]],
                           where the keys are memory region names and the values are arrays of
                           initialization data.

                           For now, we also support input of type Dict[MemoryReference, Any],
                           but this is deprecated and will be removed in a future release.
        :return: Either a float or array floats depending on ``pauli_terms``.
        """

        is_pauli_sum = False
        if isinstance(pauli_terms, PauliSum):
            progs, coeffs = pauli_terms.get_programs()
            is_pauli_sum = True
        else:
            coeffs = np.array([pt.coefficient for pt in pauli_terms])
            progs = [pt.program for pt in pauli_terms]

        if memory_map is not None:
            prep_prog = self.augment_program_with_memory_values(prep_prog, memory_map)

        bare_results = self.connection._expectation(prep_prog, progs, random_seed=self.random_seed)
        results = coeffs * bare_results
        if is_pauli_sum:
            return np.sum(results)
        return results

    @_record_call
    def run_and_measure(
        self,
        quil_program: Program,
        qubits: Optional[List[int]] = None,
        trials: int = 1,
        memory_map: Optional[
            Union[Dict[str, List[Union[int, float]]], Dict[MemoryReference, Any]]
        ] = None,
    ) -> np.ndarray:
        """
        Run a Quil program once to determine the final wavefunction, and measure multiple times.

        Alternatively, consider using ``wavefunction`` and calling ``sample_bitstrings`` on the
        resulting object.

        For a large wavefunction and a low-medium number of trials, use this function.
        On the other hand, if you're sampling a small system many times you might want to
        use ``Wavefunction.sample_bitstrings``.

        .. note:: If your program contains measurements or noisy gates, this method may not do what
            you want. If the execution of ``quil_program`` is **non-deterministic** then the
            final wavefunction from which the returned bitstrings are sampled itself only
            represents a stochastically generated sample and the outcomes sampled from
            *different* ``run_and_measure`` calls *generally sample different bitstring
            distributions*.

        :param quil_program: The program to run and measure
        :param qubits: An optional list of qubits to measure. The order of this list is
            respected in the returned bitstrings. If not provided, all qubits used in
            the program will be measured and returned in their sorted order.
        :param int trials: Number of times to sample from the prepared wavefunction.
        :param memory_map: An assignment of classical registers to values, representing an initial
                           state for the QAM's classical memory.

                           This is expected to be of type Dict[str, List[Union[int, float]]],
                           where the keys are memory region names and the values are arrays of
                           initialization data.

                           For now, we also support input of type Dict[MemoryReference, Any],
                           but this is deprecated and will be removed in a future release.
        :return: An array of measurement results (0 or 1) of shape (trials, len(qubits))
        """
        if qubits is None:
            qubits = sorted(cast(Set[int], quil_program.get_qubits(indices=True)))

        if memory_map is not None:
            quil_program = self.augment_program_with_memory_values(quil_program, memory_map)

        return self.connection._run_and_measure(
            quil_program=quil_program, qubits=qubits, trials=trials, random_seed=self.random_seed
        )

    @staticmethod
    def augment_program_with_memory_values(
        quil_program: Program,
        memory_map: Union[Dict[str, List[Union[int, float]]], Dict[MemoryReference, Any]],
    ) -> Program:
        p = Program()

        # we stupidly allowed memory_map to be of type Dict[MemoryReference, Any], whereas qc.run
        # takes a memory initialization argument of type Dict[str, List[Union[int, float]]. until
        # we are in a position to remove this, we support both styles of input.

        if len(memory_map.keys()) == 0:
            return quil_program
        elif isinstance(list(memory_map.keys())[0], MemoryReference):
            warn(
                "Use of memory_map values of type Dict[MemoryReference, Any] have been "
                "deprecated.  Please use Dict[str, List[Union[int, float]]], as with "
                "QuantumComputer.run ."
            )
            for k, v in memory_map.items():
                p += MOVE(k, v)
        elif isinstance(list(memory_map.keys())[0], str):
            for name, arr in memory_map.items():
                for index, value in enumerate(arr):
                    p += MOVE(MemoryReference(cast(str, name), offset=index), value)
        else:
            raise TypeError("Bad memory_map type; expected Dict[str, List[Union[int, float]]].")

        p += quil_program

        return percolate_declares(p)



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import networkx as nx
import numpy as np

from pyquil.device._isa import Edge, GateInfo, ISA, MeasureInfo, Qubit, isa_from_graph, isa_to_graph
from pyquil.device._specs import Specs, specs_from_graph
from pyquil.noise import NoiseModel

PERFECT_FIDELITY = 1e0
PERFECT_DURATION = 1 / 100
DEFAULT_CZ_DURATION = 200
DEFAULT_CZ_FIDELITY = 0.89
DEFAULT_ISWAP_DURATION = 200
DEFAULT_ISWAP_FIDELITY = 0.90
DEFAULT_CPHASE_DURATION = 200
DEFAULT_CPHASE_FIDELITY = 0.85
DEFAULT_XY_DURATION = 200
DEFAULT_XY_FIDELITY = 0.86
DEFAULT_RX_DURATION = 50
DEFAULT_RX_FIDELITY = 0.95
DEFAULT_MEASURE_FIDELITY = 0.90
DEFAULT_MEASURE_DURATION = 2000


class AbstractDevice(ABC):
    @abstractmethod
    def qubits(self) -> List[int]:
        """
        A sorted list of qubits in the device topology.
        """

    @abstractmethod
    def qubit_topology(self) -> nx.Graph:
        """
        The connectivity of qubits in this device given as a NetworkX graph.
        """

    @abstractmethod
    def get_isa(self, oneq_type: str = "Xhalves", twoq_type: str = "CZ") -> ISA:
        """
        Construct an ISA suitable for targeting by compilation.

        This will raise an exception if the requested ISA is not supported by the device.

        :param oneq_type: The family of one-qubit gates to target
        :param twoq_type: The family of two-qubit gates to target
        """

    @abstractmethod
    def get_specs(self) -> Optional[Specs]:
        """
        Construct a Specs object required by compilation
        """


class Device(AbstractDevice):
    """
    A device (quantum chip) that can accept programs.

    Only devices that are online will actively be
    accepting new programs. In addition to the ``self._raw`` attribute, two other attributes are
    optionally constructed from the entries in ``self._raw`` -- ``isa`` and ``noise_model`` -- which
    should conform to the dictionary format required by the ``.from_dict()`` methods for ``ISA``
    and ``NoiseModel``, respectively.

    :ivar dict _raw: Raw JSON response from the server with additional information about the device.
    :ivar ISA isa: The instruction set architecture (ISA) for the device.
    :ivar NoiseModel noise_model: The noise model for the device.
    """

    def __init__(self, name: str, raw: Dict[str, Any]):
        """
        :param name: name of the device
        :param raw: raw JSON response from the server with additional information about this device.
        """
        self.name = name
        self._raw = raw

        # TODO: Introduce distinction between supported ISAs and target ISA
        self._isa = ISA.from_dict(raw["isa"]) if "isa" in raw and raw["isa"] != {} else None
        self.specs = Specs.from_dict(raw["specs"]) if raw.get("specs") else None
        self.noise_model = (
            NoiseModel.from_dict(raw["noise_model"]) if raw.get("noise_model") else None
        )

    @property
    def isa(self) -> Optional[ISA]:
        warnings.warn("Accessing the static ISA is deprecated. Use `get_isa`", DeprecationWarning)
        return self._isa

    def qubits(self) -> List[int]:
        assert self._isa is not None
        return sorted(q.id for q in self._isa.qubits if not q.dead)

    def qubit_topology(self) -> nx.Graph:
        """
        The connectivity of qubits in this device given as a NetworkX graph.
        """
        assert self._isa is not None
        return isa_to_graph(self._isa)

    def get_specs(self) -> Optional[Specs]:
        return self.specs

    def get_isa(self, oneq_type: Optional[str] = None, twoq_type: Optional[str] = None) -> ISA:
        """
        Construct an ISA suitable for targeting by compilation.

        This will raise an exception if the requested ISA is not supported by the device.
        """
        if oneq_type is not None or twoq_type is not None:
            raise ValueError(
                "oneq_type and twoq_type are both fatally deprecated. If you want to "
                "make an ISA with custom gate types, you'll have to do it by hand."
            )

        def safely_get(attr: str, index: Union[int, Tuple[int, ...]], default: Any) -> Any:
            if self.specs is None:
                return default

            getter = getattr(self.specs, attr, None)
            if getter is None:
                return default

            array = getter()
            if (isinstance(index, int) and index < len(array)) or index in array:
                return array[index]
            else:
                return default

        def qubit_type_to_gates(q: Qubit) -> List[Union[GateInfo, MeasureInfo]]:
            gates: List[Union[GateInfo, MeasureInfo]] = [
                MeasureInfo(
                    operator="MEASURE",
                    qubit=q.id,
                    target="_",
                    fidelity=safely_get("fROs", q.id, DEFAULT_MEASURE_FIDELITY),
                    duration=DEFAULT_MEASURE_DURATION,
                ),
                MeasureInfo(
                    operator="MEASURE",
                    qubit=q.id,
                    target=None,
                    fidelity=safely_get("fROs", q.id, DEFAULT_MEASURE_FIDELITY),
                    duration=DEFAULT_MEASURE_DURATION,
                ),
            ]
            if q.type is None or "Xhalves" in q.type:
                gates += [
                    GateInfo(
                        operator="RZ",
                        parameters=["_"],
                        arguments=[q.id],
                        duration=PERFECT_DURATION,
                        fidelity=PERFECT_FIDELITY,
                    ),
                    GateInfo(
                        operator="RX",
                        parameters=[0.0],
                        arguments=[q.id],
                        duration=DEFAULT_RX_DURATION,
                        fidelity=PERFECT_FIDELITY,
                    ),
                ]
                gates += [
                    GateInfo(
                        operator="RX",
                        parameters=[param],
                        arguments=[q.id],
                        duration=DEFAULT_RX_DURATION,
                        fidelity=safely_get("f1QRBs", q.id, DEFAULT_RX_FIDELITY),
                    )
                    for param in [np.pi, -np.pi, np.pi / 2, -np.pi / 2]
                ]
            if q.type is not None and "WILDCARD" in q.type:
                gates += [
                    GateInfo(
                        operator="_",
                        parameters="_",
                        arguments=[q.id],
                        duration=PERFECT_DURATION,
                        fidelity=PERFECT_FIDELITY,
                    )
                ]
            return gates

        def edge_type_to_gates(e: Edge) -> List[GateInfo]:
            gates: List[GateInfo] = []
            if (
                e is None
                or isinstance(e.type, str)
                and "CZ" == e.type
                or isinstance(e.type, list)
                and "CZ" in e.type
            ):
                gates += [
                    GateInfo(
                        operator="CZ",
                        parameters=[],
                        arguments=["_", "_"],
                        duration=DEFAULT_CZ_DURATION,
                        fidelity=safely_get("fCZs", tuple(e.targets), DEFAULT_CZ_FIDELITY),
                    )
                ]
            if (
                e is None
                or isinstance(e.type, str)
                and "ISWAP" == e.type
                or isinstance(e.type, list)
                and "ISWAP" in e.type
            ):
                gates += [
                    GateInfo(
                        operator="ISWAP",
                        parameters=[],
                        arguments=["_", "_"],
                        duration=DEFAULT_ISWAP_DURATION,
                        fidelity=safely_get("fISWAPs", tuple(e.targets), DEFAULT_ISWAP_FIDELITY),
                    )
                ]
            if (
                e is None
                or isinstance(e.type, str)
                and "CPHASE" == e.type
                or isinstance(e.type, list)
                and "CPHASE" in e.type
            ):
                gates += [
                    GateInfo(
                        operator="CPHASE",
                        parameters=["theta"],
                        arguments=["_", "_"],
                        duration=DEFAULT_CPHASE_DURATION,
                        fidelity=safely_get("fCPHASEs", tuple(e.targets), DEFAULT_CPHASE_FIDELITY),
                    )
                ]
            if (
                e is None
                or isinstance(e.type, str)
                and "XY" == e.type
                or isinstance(e.type, list)
                and "XY" in e.type
            ):
                gates += [
                    GateInfo(
                        operator="XY",
                        parameters=["theta"],
                        arguments=["_", "_"],
                        duration=DEFAULT_XY_DURATION,
                        fidelity=safely_get("fXYs", tuple(e.targets), DEFAULT_XY_FIDELITY),
                    )
                ]
            if (
                e is None
                or isinstance(e.type, str)
                and "WILDCARD" == e.type
                or isinstance(e.type, list)
                and "WILDCARD" in e.type
            ):
                gates += [
                    GateInfo(
                        operator="_",
                        parameters="_",
                        arguments=["_", "_"],
                        duration=PERFECT_DURATION,
                        fidelity=PERFECT_FIDELITY,
                    )
                ]
            return gates

        assert self._isa is not None
        qubits = [
            Qubit(id=q.id, type=None, dead=q.dead, gates=qubit_type_to_gates(q))
            for q in self._isa.qubits
        ]
        edges = [
            Edge(targets=e.targets, type=None, dead=e.dead, gates=edge_type_to_gates(e))
            for e in self._isa.edges
        ]
        return ISA(qubits, edges)

    def __str__(self) -> str:
        return "<Device {}>".format(self.name)

    def __repr__(self) -> str:
        return str(self)


class NxDevice(AbstractDevice):
    """A shim over the AbstractDevice API backed by a NetworkX graph.

    A ``Device`` holds information about the physical device.
    Specifically, you might want to know about connectivity, available gates, performance specs,
    and more. This class implements the AbstractDevice API for devices not available via
    ``get_devices()``. Instead, the user is responsible for constructing a NetworkX
    graph which represents a chip topology.
    """

    def __init__(self, topology: nx.Graph) -> None:
        self.topology = topology

    def qubit_topology(self) -> nx.Graph:
        return self.topology

    def get_isa(
        self, oneq_type: str = "Xhalves", twoq_type: Optional[Union[str, List[str]]] = None
    ) -> ISA:
        return isa_from_graph(self.topology, oneq_type=oneq_type, twoq_type=twoq_type)

    def get_specs(self) -> Specs:
        return specs_from_graph(self.topology)

    def qubits(self) -> List[int]:
        return sorted(self.topology.nodes)

    def edges(self) -> List[Tuple[Any, ...]]:
        return sorted(tuple(sorted(pair)) for pair in self.topology.edges)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from collections import defaultdict
import uuid
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
from rpcq._client import Client, ClientAuthConfig
from rpcq.messages import BinaryExecutableResponse, QPURequest, ParameterAref

from pyquil.parser import parse
from pyquil.api._base_connection import Engagement, ForestSession
from pyquil.api._error_reporting import _record_call
from pyquil.api._errors import UserMessageError
from pyquil.api._logger import logger
from pyquil.api._qam import QAM
from pyquil.quil import Program
from pyquil.quilatom import MemoryReference, BinaryExp, Function, Parameter, ExpressionDesignator
from pyquil.quilbase import Gate


def decode_buffer(buffer: Dict[str, Any]) -> np.ndarray:
    """
    Translate a DataBuffer into a numpy array.

    :param buffer: Dictionary with 'data' byte array, 'dtype', and 'shape' fields
    :return: NumPy array of decoded data
    """
    buf = np.frombuffer(buffer["data"], dtype=buffer["dtype"])
    return buf.reshape(buffer["shape"])


def _extract_bitstrings(
    ro_sources: List[Optional[Tuple[int, int]]], buffers: Dict[str, np.ndarray]
) -> np.ndarray:
    """
    De-mux qubit readout results and assemble them into the ro-bitstrings in the correct order.

    :param ro_sources: Specification of the ro_sources, cf
        :py:func:`pyquil.api._compiler._collect_classical_memory_write_locations`.
        It is a list whose value ``(q, m)`` at index ``addr`` records that the ``m``-th measurement
        of qubit ``q`` was measured into ``ro`` address ``addr``. A value of `None` means nothing
        was measured into ``ro`` address ``addr``.
    :param buffers: A dictionary of readout results returned from the qpu.
    :return: A numpy array of shape ``(num_shots, len(ro_sources))`` with the readout bits.
    """
    # hack to extract num_shots indirectly from the shape of the returned data
    first, *rest = buffers.values()
    num_shots = first.shape[0]
    bitstrings = np.zeros((num_shots, len(ro_sources)), dtype=np.int64)
    for col_idx, src in enumerate(ro_sources):
        if src:
            qubit, meas_idx = src
            buf = buffers[f"q{qubit}"]
            if buf.ndim == 1:
                buf = buf.reshape((num_shots, 1))
            bitstrings[:, col_idx] = buf[:, meas_idx]
    return bitstrings


class QPU(QAM):
    @_record_call
    def __init__(
        self,
        endpoint: Optional[str] = None,
        user: str = "pyquil-user",
        priority: int = 1,
        *,
        session: Optional[ForestSession] = None,
    ) -> None:
        """
        A connection to the QPU.

        :param endpoint: Address to connect to the QPU server. If not provided, the
                         endpoint provided by engagement with dispatch is used. One or both must be
                         available and valid.
        :param user: A string identifying who's running jobs.
        :param priority: The priority with which to insert jobs into the QPU queue. Lower
                         integers correspond to higher priority.
        :param session: ForestSession object, which manages engagement and configuration.
        """
        if not (session or endpoint):
            raise ValueError("QPU requires either `session` or `endpoint`.")

        self.session = session

        self.endpoint = endpoint
        self.priority = priority
        self.user = user
        self._client: Optional[Client] = None
        self._client_engagement: Optional[Engagement] = None
        self._last_results: Dict[str, np.ndarray] = {}

        super().__init__()

    def _build_client(self) -> Client:
        endpoint: Optional[str] = None
        if self.endpoint:
            endpoint = self.endpoint
        elif self.session and self.session.config.qpu_url:
            endpoint = self.session.config.qpu_url

        if endpoint is None:
            raise UserMessageError(
                """It looks like you've tried to run a program against a QPU but do
not currently have a reservation on one. To reserve time on Rigetti
QPUs, use the command line interface, qcs, which comes pre-installed
in your QMI. From within your QMI, type:

    qcs reserve --lattice <lattice-name>

For more information, please see the docs at
https://www.rigetti.com/qcs/docs/reservations or reach out to Rigetti
support at support@rigetti.com."""
            )

        logger.debug("QPU Client connecting to %s", endpoint)

        return Client(endpoint, auth_config=self._get_client_auth_config())

    @property
    def client(self) -> Client:
        """
        Return a memoized client with a valid engagement if it exists; otherwise, build a new one
        and return it. If this QPU object does not have a ForestSession, then engagement is not
        applicable.
        """
        engagement_is_valid = self._client_engagement and self._client_engagement.is_valid()
        if not (self._client and (self.session is None or engagement_is_valid)):
            self._client = self._build_client()
        return self._client

    def _get_client_auth_config(self) -> Optional[ClientAuthConfig]:
        if self.session and self.session.config.engagement is not None:
            # We store the engagement used to construct this client so that we can later check
            # for validity
            self._client_engagement = self.session.config.engagement
            return ClientAuthConfig(
                client_public_key=self._client_engagement.client_public_key,
                client_secret_key=self._client_engagement.client_secret_key,
                server_public_key=self._client_engagement.server_public_key,
            )
        return None

    def get_version_info(self) -> Dict[str, Any]:
        """
        Return version information for this QPU's execution engine and its dependencies.

        :return: Dictionary of version information.
        """
        return cast(Dict[str, Any], self.client.call("get_version_info"))

    @_record_call
    def load(self, executable: BinaryExecutableResponse) -> "QPU":
        """
        Initialize a QAM into a fresh state. Load the executable and parse the expressions
        in the recalculation table (if any) into pyQuil Expression objects.

        :param executable: Load a compiled executable onto the QAM.
        """
        super().load(executable)
        if hasattr(self._executable, "recalculation_table"):
            recalculation_table = self._executable.recalculation_table  # type: ignore
            for memory_reference, recalc_rule in recalculation_table.items():
                # We can only parse complete lines of Quil, so we wrap the arithmetic expression
                # in a valid Quil instruction to parse it.
                # TODO: This hack should be replaced after #687
                expression = cast(Gate, parse(f"RZ({recalc_rule}) 0")[0]).params[0]
                recalculation_table[memory_reference] = expression
        return self

    @_record_call
    def run(self, run_priority: Optional[int] = None) -> "QPU":
        """
        Run a pyquil program on the QPU.

        This formats the classified data from the QPU server by stacking measured bits into
        an array of shape (trials, classical_addresses). The mapping of qubit to
        classical address is backed out from MEASURE instructions in the program, so
        only do measurements where there is a 1-to-1 mapping between qubits and classical
        addresses.

        :param run_priority: The priority with which to insert jobs into the QPU queue. Lower
                             integers correspond to higher priority. If not specified, the QPU
                             object's default priority is used.
        :return: The QPU object itself.
        """
        # This prevents a common error where users expect QVM.run()
        # and QPU.run() to be interchangeable. QPU.run() needs the
        # supplied executable to have been compiled, QVM.run() does not.
        if isinstance(self._executable, Program):
            raise TypeError(
                "It looks like you have provided a Program where an Executable"
                " is expected. Please use QuantumComputer.compile() to compile"
                " your program."
            )
        super().run()

        assert self._executable is not None
        request = QPURequest(
            program=self._executable.program,
            patch_values=self._build_patch_values(),
            id=str(uuid.uuid4()),
        )

        job_priority = run_priority if run_priority is not None else self.priority
        job_id = self.client.call(
            "execute_qpu_request", request=request, user=self.user, priority=job_priority
        )
        results = self._get_buffers(job_id)
        ro_sources = self._executable.ro_sources  # type: ignore

        if results:
            bitstrings = _extract_bitstrings(ro_sources, results)
        elif not ro_sources:
            warnings.warn(
                "You are running a QPU program with no MEASURE instructions. "
                "The result of this program will always be an empty array. Are "
                "you sure you didn't mean to measure some of your qubits?"
            )
            bitstrings = np.zeros((0, 0), dtype=np.int64)
        else:
            bitstrings = None

        self._memory_results = defaultdict(lambda: None)
        self._memory_results["ro"] = bitstrings
        self._last_results = results

        return self

    def _get_buffers(self, job_id: str) -> Dict[str, np.ndarray]:
        """
        Return the decoded result buffers for particular job_id.

        :param job_id: Unique identifier for the job in question
        :return: Decoded buffers or throw an error
        """
        buffers = self.client.call("get_buffers", job_id, wait=True)
        return {k: decode_buffer(v) for k, v in buffers.items()}

    def _build_patch_values(self) -> Dict[str, List[Union[int, float]]]:
        patch_values = {}

        # Now that we are about to run, we have to resolve any gate parameter arithmetic that was
        # saved in the executable's recalculation table, and add those values to the variables shim
        self._update_variables_shim_with_recalculation_table()

        # Initialize our patch table
        if hasattr(self._executable, "recalculation_table"):
            memory_ref_names = list(
                set(mr.name for mr in self._executable.recalculation_table.keys())  # type: ignore
            )
            if memory_ref_names != []:
                assert len(memory_ref_names) == 1, (
                    "We expected only one declared memory region for "
                    "the gate parameter arithmetic replacement references."
                )
                memory_reference_name = memory_ref_names[0]
                patch_values[memory_reference_name] = [0.0] * len(
                    self._executable.recalculation_table  # type: ignore
                )

        assert isinstance(self._executable, BinaryExecutableResponse)
        for name, spec in self._executable.memory_descriptors.items():
            # NOTE: right now we fake reading out measurement values into classical memory
            if name == "ro":
                continue
            initial_value = 0.0 if spec.type == "REAL" else 0
            patch_values[name] = [initial_value] * spec.length

        # Fill in our patch table
        for k, v in self._variables_shim.items():
            # NOTE: right now we fake reading out measurement values into classical memory
            if k.name == "ro":
                continue

            # floats stored in tsunami memory are expected to be in revolutions rather than radians.
            if isinstance(v, float):
                v /= 2 * np.pi

            patch_values[k.name][k.index] = v

        return patch_values

    def _update_variables_shim_with_recalculation_table(self) -> None:
        """
        Update self._variables_shim with the final values to be patched into the gate parameters,
        according to the arithmetic expressions in the original program.

        For example::

            DECLARE theta REAL
            DECLARE beta REAL
            RZ(3 * theta) 0
            RZ(beta+theta) 0

        gets translated to::

            DECLARE theta REAL
            DECLARE __P REAL[2]
            RZ(__P[0]) 0
            RZ(__P[1]) 0

        and the recalculation table will contain::

            {
                ParameterAref('__P', 0): Mul(3.0, <MemoryReference theta[0]>),
                ParameterAref('__P', 1): Add(<MemoryReference beta[0]>, <MemoryReference theta[0]>)
            }

        Let's say we've made the following two function calls:

        .. code-block:: python

            qpu.write_memory(region_name='theta', value=0.5)
            qpu.write_memory(region_name='beta', value=0.1)

        After executing this function, our self.variables_shim in the above example would contain
        the following:

        .. code-block:: python

            {
                ParameterAref('theta', 0): 0.5,
                ParameterAref('beta', 0): 0.1,
                ParameterAref('__P', 0): 1.5,       # (3.0) * theta[0]
                ParameterAref('__P', 1): 0.6        # beta[0] + theta[0]
            }

        Once the _variables_shim is filled, execution continues as with regular binary patching.
        """
        if not hasattr(self._executable, "recalculation_table"):
            # No recalculation table, no work to be done here.
            return
        for mref, expression in self._executable.recalculation_table.items():  # type: ignore
            # Replace the user-declared memory references with any values the user has written,
            # coerced to a float because that is how we declared it.
            self._variables_shim[mref] = float(self._resolve_memory_references(expression))

    def _resolve_memory_references(self, expression: ExpressionDesignator) -> Union[float, int]:
        """
        Traverse the given Expression, and replace any Memory References with whatever values
        have been so far provided by the user for those memory spaces. Declared memory defaults
        to zero.

        :param expression: an Expression
        """
        if isinstance(expression, BinaryExp):
            left = self._resolve_memory_references(expression.op1)
            right = self._resolve_memory_references(expression.op2)
            return cast(Union[float, int], expression.fn(left, right))
        elif isinstance(expression, Function):
            return cast(
                Union[float, int],
                expression.fn(self._resolve_memory_references(expression.expression)),
            )
        elif isinstance(expression, Parameter):
            raise ValueError(f"Unexpected Parameter in gate expression: {expression}")
        elif isinstance(expression, (float, int)):
            return expression
        elif isinstance(expression, MemoryReference):
            return self._variables_shim.get(
                ParameterAref(name=expression.name, index=expression.offset), 0
            )
        else:
            raise ValueError(f"Unexpected expression in gate parameter: {expression}")

    @_record_call
    def reset(self) -> None:
        """
        Reset the state of the underlying QAM, and the QPU Client connection.
        """
        super().reset()

        self._client = None



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import logging
import sys
import warnings
from requests.exceptions import RequestException
from typing import Dict, Any, List, Optional, Tuple, Union, cast
from collections import Counter

from rpcq._base import Message, to_json, from_json
from rpcq._client import Client
from rpcq.messages import (
    BinaryExecutableRequest,
    BinaryExecutableResponse,
    NativeQuilRequest,
    TargetDevice,
    PyQuilExecutableResponse,
    ParameterSpec,
    RewriteArithmeticRequest,
    RewriteArithmeticResponse,
)
from urllib.parse import urljoin

from pyquil.api._base_connection import ForestSession
from pyquil.api._qac import AbstractCompiler
from pyquil.api._error_reporting import _record_call
from pyquil.api._errors import UserMessageError
from pyquil.device._main import AbstractDevice, Device
from pyquil.parser import parse_program
from pyquil.quil import Program
from pyquil.quilbase import Measurement, Declare
from pyquil.version import __version__

if sys.version_info < (3, 7):
    from pyquil.external.dataclasses import dataclass
else:
    from dataclasses import dataclass


_log = logging.getLogger(__name__)

PYQUIL_PROGRAM_PROPERTIES = ["native_quil_metadata", "num_shots"]


class QuilcVersionMismatch(Exception):
    pass


class QuilcNotRunning(Exception):
    pass


class QPUCompilerNotRunning(Exception):
    pass


def check_quilc_version(version_dict: Dict[str, str]) -> None:
    """
    Verify that there is no mismatch between pyquil and quilc versions.

    :param version_dict: Dictionary containing version information about quilc.
    """
    quilc_version = version_dict["quilc"]
    major, minor, patch = map(int, quilc_version.split("."))
    if major == 1 and minor < 8:
        raise QuilcVersionMismatch(
            "Must use quilc >= 1.8.0 with pyquil >= 2.8.0, but you "
            f"have quilc {quilc_version} and pyquil {__version__}"
        )


def _extract_attribute_dictionary_from_program(program: Program) -> Dict[str, Any]:
    """
    Collects the attributes from PYQUIL_PROGRAM_PROPERTIES on the Program object program
    into a dictionary.

    :param program: Program to collect attributes from.
    :return: Dictionary of attributes, keyed on the string attribute name.
    """
    attrs = {}
    for prop in PYQUIL_PROGRAM_PROPERTIES:
        attrs[prop] = getattr(program, prop)
    return attrs


def _extract_program_from_pyquil_executable_response(response: PyQuilExecutableResponse) -> Program:
    """
    Unpacks a rpcq PyQuilExecutableResponse object into a pyQuil Program object.

    :param response: PyQuilExecutableResponse object to be unpacked.
    :return: Resulting pyQuil Program object.
    """
    p = Program(response.program)
    for attr, val in response.attributes.items():
        setattr(p, attr, val)
    return p


def _collect_classical_memory_write_locations(program: Program) -> List[Optional[Tuple[int, int]]]:
    """Collect classical memory locations that are the destination of MEASURE instructions

    These locations are important for munging output buffers returned from the QPU
    server to the shape expected by the user.

    This is secretly stored on BinaryExecutableResponse. We're careful to make sure
    these objects are json serializable.

    :return: list whose value `(q, m)` at index `addr` records that the `m`-th measurement of
        qubit `q` was measured into `ro` address `addr`. A value of `None` means nothing was
        measured into `ro` address `addr`.
    """
    ro_size = None
    for instr in program:
        if isinstance(instr, Declare) and instr.name == "ro":
            if ro_size is not None:
                raise ValueError(
                    "I found multiple places where a register named `ro` is declared! "
                    "Please only declare one register named `ro`."
                )
            ro_size = instr.memory_size

    measures_by_qubit: Dict[int, int] = Counter()
    ro_sources: Dict[int, Tuple[int, int]] = {}

    for instr in program:
        if isinstance(instr, Measurement):
            q = instr.qubit.index
            if instr.classical_reg:
                offset = instr.classical_reg.offset
                assert instr.classical_reg.name == "ro", instr.classical_reg.name
                if offset in ro_sources:
                    _log.warning(
                        f"Overwriting the measured result in register "
                        f"{instr.classical_reg} from qubit {ro_sources[offset]} "
                        f"to qubit {q}"
                    )
                # we track how often each qubit is measured (per shot) and into which register it is
                # measured in its n-th measurement.
                ro_sources[offset] = (q, measures_by_qubit[q])
            measures_by_qubit[q] += 1
    if ro_size:
        return [ro_sources.get(i) for i in range(ro_size)]
    elif ro_sources:
        raise ValueError(
            "Found MEASURE instructions, but no 'ro' or 'ro_table' region was declared."
        )
    else:
        return []


def _collect_memory_descriptors(program: Program) -> Dict[str, ParameterSpec]:
    """Collect Declare instructions that are important for building the patch table.

    This is secretly stored on BinaryExecutableResponse. We're careful to make sure
    these objects are json serializable.

    :return: A dictionary of variable names to specs about the declared region.
    """
    return {
        instr.name: ParameterSpec(type=instr.memory_type, length=instr.memory_size)
        for instr in program
        if isinstance(instr, Declare)
    }


class QPUCompiler(AbstractCompiler):
    @_record_call
    def __init__(
        self,
        quilc_endpoint: Optional[str],
        qpu_compiler_endpoint: Optional[str],
        device: AbstractDevice,
        timeout: int = 10,
        name: Optional[str] = None,
        *,
        session: Optional[ForestSession] = None,
    ) -> None:
        """
        Client to communicate with the Compiler Server.

        :param quilc_endpoint: TCP or IPC endpoint of the Quil Compiler (quilc).
        :param qpu_compiler_endpoint: TCP or IPC endpoint of the QPU Compiler.
        :param device: PyQuil Device object to use as compilation target.
        :param timeout: Number of seconds to wait for a response from the client.
        :param name: Name of the lattice being targeted.
        :param session: ForestSession object, which manages engagement and configuration.
        """

        if not (session or (quilc_endpoint and qpu_compiler_endpoint)):
            raise ValueError(
                "QPUCompiler requires either `session` or both of `quilc_endpoint` and "
                "`qpu_compiler_endpoint`."
            )

        self.session = session
        self.timeout = timeout

        if quilc_endpoint:
            _quilc_endpoint = quilc_endpoint
        elif self.session and self.session.config.quilc_url:
            _quilc_endpoint = self.session.config.quilc_url
        else:
            raise ValueError("Must provide a 'quilc_endpoint' or a 'session'")

        if not _quilc_endpoint.startswith("tcp://"):
            raise ValueError(
                f"PyQuil versions >= 2.4 can only talk to quilc "
                f"versions >= 1.4 over network RPCQ.  You've supplied the "
                f"endpoint '{quilc_endpoint}', but this doesn't look like a network "
                f"ZeroMQ address, which has the form 'tcp://domain:port'. "
                f"You might try clearing (or correcting) your COMPILER_URL "
                f"environment variable and removing (or correcting) the "
                f"compiler_server_address line from your .forest_config file."
            )

        self.quilc_client = Client(_quilc_endpoint, timeout=timeout)

        self.qpu_compiler_endpoint = qpu_compiler_endpoint
        self._qpu_compiler_client: Optional[Union[Client, HTTPCompilerClient]] = None

        self._device = device
        td = TargetDevice(isa=device.get_isa().to_dict(), specs=None)  # type: ignore
        self.target_device = td
        self.name = name

        try:
            self.connect()
        except QuilcNotRunning as e:
            warnings.warn(f"{e}. Compilation using quilc will not be available.")
        except QPUCompilerNotRunning as e:
            warnings.warn(f"{e}. Compilation using the QPU compiler will not be available.")

    @property
    def qpu_compiler_client(self) -> Optional[Union[Client, "HTTPCompilerClient"]]:
        if not self._qpu_compiler_client:
            endpoint: Optional[str] = None
            if self.qpu_compiler_endpoint:
                endpoint = self.qpu_compiler_endpoint
            elif self.session:
                endpoint = self.session.config.qpu_compiler_url

            if endpoint is not None:
                if endpoint.startswith(("http://", "https://")):
                    assert isinstance(self._device, Device) and self.session is not None
                    device_endpoint = urljoin(
                        endpoint, f'devices/{self._device._raw["device_name"]}/'
                    )
                    self._qpu_compiler_client = HTTPCompilerClient(
                        endpoint=device_endpoint, session=self.session
                    )
                elif endpoint.startswith("tcp://"):
                    self._qpu_compiler_client = Client(endpoint, timeout=self.timeout)
                else:
                    raise UserMessageError(
                        "Invalid endpoint provided to QPUCompiler. Expected protocol in [http://, "
                        f"https://, tcp://], but received endpoint {endpoint}"
                    )

        assert (
            isinstance(self._qpu_compiler_client, (Client, HTTPCompilerClient))
            or self._qpu_compiler_client is None
        )
        return self._qpu_compiler_client

    def connect(self) -> None:
        self._connect_quilc()
        if self.qpu_compiler_client:
            self._connect_qpu_compiler()

    def _connect_quilc(self) -> None:
        try:
            quilc_version_dict = self.quilc_client.call("get_version_info", rpc_timeout=1)
            check_quilc_version(quilc_version_dict)
        except TimeoutError:
            raise QuilcNotRunning(f"No quilc server reachable at {self.quilc_client.endpoint}")

    def _connect_qpu_compiler(self) -> None:
        assert self.qpu_compiler_client is not None
        try:
            self.qpu_compiler_client.call("get_version_info", rpc_timeout=1)
        except TimeoutError:
            raise QPUCompilerNotRunning(
                f"No QPU compiler server reachable at {self.qpu_compiler_client.endpoint}"
            )

    def get_version_info(self) -> Dict[str, Any]:
        quilc_version_info = self.quilc_client.call("get_version_info", rpc_timeout=1)
        if self.qpu_compiler_client:
            qpu_compiler_version_info = self.qpu_compiler_client.call(
                "get_version_info", rpc_timeout=1
            )
            return {"quilc": quilc_version_info, "qpu_compiler": qpu_compiler_version_info}
        return {"quilc": quilc_version_info}

    @_record_call
    def quil_to_native_quil(self, program: Program, *, protoquil: Optional[bool] = None) -> Program:
        self._connect_quilc()
        request = NativeQuilRequest(quil=program.out(), target_device=self.target_device)
        response = self.quilc_client.call(
            "quil_to_native_quil", request, protoquil=protoquil
        ).asdict()
        nq_program = parse_program(response["quil"])
        nq_program.native_quil_metadata = response["metadata"]
        nq_program.num_shots = program.num_shots
        return nq_program

    @_record_call
    def native_quil_to_executable(self, nq_program: Program) -> Optional[BinaryExecutableResponse]:
        if not self.qpu_compiler_client:
            raise UserMessageError(
                "It looks like you're trying to compile to an executable, but "
                "do not have access to the QPU compiler endpoint. Make sure you "
                "are engaged to the QPU before trying to do this."
            )

        self._connect_qpu_compiler()

        if nq_program.native_quil_metadata is None:
            warnings.warn(
                "It looks like you're trying to call `native_quil_to_binary` on a "
                "Program that hasn't been compiled via `quil_to_native_quil`. This is "
                "ok if you've hand-compiled your program to our native gateset, "
                "but be careful!"
            )

        arithmetic_request = RewriteArithmeticRequest(quil=nq_program.out())
        arithmetic_response: RewriteArithmeticResponse = cast(
            RewriteArithmeticResponse,
            self.quilc_client.call("rewrite_arithmetic", arithmetic_request),
        )

        request = BinaryExecutableRequest(
            quil=arithmetic_response.quil, num_shots=nq_program.num_shots
        )
        response: BinaryExecutableResponse = cast(
            BinaryExecutableResponse,
            self.qpu_compiler_client.call("native_quil_to_binary", request),
        )

        # hack! we're storing a little extra info in the executable binary that we don't want to
        # expose to anyone outside of our own private lives: not the user, not the Forest server,
        # not anyone.
        response.recalculation_table = arithmetic_response.recalculation_table  # type: ignore
        response.memory_descriptors = _collect_memory_descriptors(nq_program)
        response.ro_sources = _collect_classical_memory_write_locations(nq_program)
        return response

    @_record_call
    def reset(self) -> None:
        """
        Reset the state of the QPUCompiler Client connections.
        """
        self._qpu_compiler_client = None


class QVMCompiler(AbstractCompiler):
    @_record_call
    def __init__(self, endpoint: str, device: AbstractDevice, timeout: float = 10) -> None:
        """
        Client to communicate with the Compiler Server.

        :param endpoint: TCP or IPC endpoint of the Compiler Server
        :param device: PyQuil Device object to use as compilation target
        """

        if not endpoint.startswith("tcp://"):
            raise ValueError(
                f"PyQuil versions >= 2.4 can only talk to quilc "
                f"versions >= 1.4 over network RPCQ.  You've supplied the "
                f"endpoint '{endpoint}', but this doesn't look like a network "
                f"ZeroMQ address, which has the form 'tcp://domain:port'. "
                f"You might try clearing (or correcting) your COMPILER_URL "
                f"environment variable and removing (or correcting) the "
                f"compiler_server_address line from your .forest_config file."
            )

        self.endpoint = endpoint
        self.client = Client(endpoint, timeout=timeout)
        td = TargetDevice(isa=device.get_isa().to_dict(), specs=None)  # type: ignore
        self.target_device = td

        try:
            self.connect()
        except QuilcNotRunning as e:
            warnings.warn(f"{e}. Compilation using quilc will not be available.")

    def connect(self) -> None:
        try:
            version_dict = self.get_version_info()
            check_quilc_version(version_dict)
        except TimeoutError:
            raise QuilcNotRunning(f"No quilc server running at {self.client.endpoint}")

    def get_version_info(self) -> Dict[str, Any]:
        return cast(Dict[str, Any], self.client.call("get_version_info", rpc_timeout=1))

    @_record_call
    def quil_to_native_quil(self, program: Program, *, protoquil: Optional[bool] = None) -> Program:
        self.connect()
        request = NativeQuilRequest(quil=program.out(), target_device=self.target_device)
        response = self.client.call("quil_to_native_quil", request, protoquil=protoquil).asdict()
        nq_program = parse_program(response["quil"])
        nq_program.native_quil_metadata = response["metadata"]
        nq_program.num_shots = program.num_shots
        return nq_program

    @_record_call
    def native_quil_to_executable(self, nq_program: Program) -> PyQuilExecutableResponse:
        return PyQuilExecutableResponse(
            program=nq_program.out(),
            attributes=_extract_attribute_dictionary_from_program(nq_program),
        )

    @_record_call
    def reset(self) -> None:
        """
        Reset the state of the QVMCompiler quilc connection.
        """
        timeout = self.client.timeout
        self.client.close()  # type: ignore
        self.client = Client(self.endpoint, timeout=timeout)


@dataclass
class HTTPCompilerClient:
    """
    A class which partially implements the interface of rpcq.Client, to allow the QPUCompiler to
    send compilation requests over HTTP(S) rather than ZeroMQ.

    :param endpoint: The base url to which rpcq methods will be appended.
    :param session: The ForestSession object which manages headers and authentication.
    """

    endpoint: str
    session: ForestSession

    def call(
        self, method: str, payload: Optional[Message] = None, *, rpc_timeout: int = 30
    ) -> Message:
        """
        A partially-compatible implementation of rpcq.Client#call, which allows calling rpcq
        methods over HTTP following the scheme:

            POST <endpoint>/<method>
            body: json-serialized <payload>

        This implementation is meant for use only with the QPUCompiler and is not intended to be
        a fully-compatible port of rpcq from ZeroMQ to HTTP.

        If the request succeeds (per HTTP response code), the body of the response is parsed into
        an RPCQ Message.

        If the request fails, the response body should be a JSON object with a ``message`` field
        indicating the cause of the failure. If present, that message is delivered to the user.

        :param payload: The rpcq message body.
        :param rpc_timeout: The number of seconds to wait to read data back from the service
            after connection.
            @see https://requests.readthedocs.io/en/master/user/advanced/#timeouts
        """
        url = urljoin(self.endpoint, method)

        if payload:
            body = to_json(payload)  # type: ignore
        else:
            body = None

        response = self.session.post(url, json=body, timeout=(1, rpc_timeout))

        try:
            response.raise_for_status()
        except RequestException as e:
            message = f"QPU Compiler {method} failed: "

            try:
                contents = response.json()
            except Exception:
                contents = None

            if contents and contents.get("message"):
                message += contents.get("message")
            else:
                message += "please try again or contact support."

            raise UserMessageError(message) from e

        return cast(Message, from_json(response.text))  # type: ignore



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import warnings
import numpy as np
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Union, cast

from requests.exceptions import ConnectionError
from rpcq.messages import PyQuilExecutableResponse

from pyquil.api._base_connection import (
    validate_qubit_list,
    validate_noise_probabilities,
    TYPE_MULTISHOT_MEASURE,
    TYPE_WAVEFUNCTION,
    TYPE_EXPECTATION,
    post_json,
    ForestConnection,
)
from pyquil.api._compiler import QVMCompiler, _extract_program_from_pyquil_executable_response
from pyquil.api._config import PyquilConfig
from pyquil.api._error_reporting import _record_call
from pyquil.api._qam import QAM
from pyquil.device._main import Device
from pyquil.gates import MOVE
from pyquil.noise import NoiseModel, apply_noise_model
from pyquil.paulis import PauliSum, PauliTerm
from pyquil.quil import Program, get_classical_addresses_from_program, percolate_declares
from pyquil.quilatom import MemoryReference
from pyquil.version import __version__
from pyquil.wavefunction import Wavefunction


class QVMVersionMismatch(Exception):
    pass


class QVMNotRunning(Exception):
    pass


def check_qvm_version(version: str) -> None:
    """
    Verify that there is no mismatch between pyquil and QVM versions.

    :param version: The version of the QVM
    """
    major, minor, patch = map(int, version.split("."))
    if major == 1 and minor < 8:
        raise QVMVersionMismatch(
            "Must use QVM >= 1.8.0 with pyquil >= 2.8.0, but you "
            f"have QVM {version} and pyquil {__version__}"
        )


class QVMConnection(object):
    """
    Represents a connection to the QVM.
    """

    @_record_call
    def __init__(
        self,
        device: Optional[Device] = None,
        endpoint: Optional[str] = None,
        gate_noise: Optional[List[float]] = None,
        measurement_noise: Optional[List[float]] = None,
        random_seed: Optional[int] = None,
        compiler_endpoint: Optional[str] = None,
    ):
        """
        Constructor for QVMConnection. Sets up any necessary security, and establishes the noise
        model to use.

        :param device: The optional device, from which noise will be added by default to all
            programs run on this instance.
        :param endpoint: The endpoint of the server for running small jobs
        :param gate_noise: A list of three numbers [Px, Py, Pz] indicating the probability of an X,
            Y, or Z gate getting applied to each qubit after a gate application or reset.
        :param measurement_noise: A list of three numbers [Px, Py, Pz] indicating the probability of
            an X, Y, or Z gate getting applied before a a measurement.
        :param random_seed: A seed for the QVM's random number generators. Either None (for an
            automatically generated seed) or a non-negative integer.
        """
        if endpoint is None:
            pyquil_config = PyquilConfig()
            endpoint = pyquil_config.qvm_url

        if compiler_endpoint is None:
            pyquil_config = PyquilConfig()
            compiler_endpoint = pyquil_config.quilc_url

        if (device is not None and device.noise_model is not None) and (
            gate_noise is not None or measurement_noise is not None
        ):
            raise ValueError(
                """
You have attempted to supply the QVM with both a device noise model
(by having supplied a device argument), as well as either gate_noise
or measurement_noise. At this time, only one may be supplied.

To read more about supplying noise to the QVM, see
http://pyquil.readthedocs.io/en/latest/noise_models.html#support-for-noisy-gates-on-the-rigetti-qvm.
"""
            )

        if device is not None and device.noise_model is None:
            warnings.warn(
                """
You have supplied the QVM with a device that does not have a noise model. No noise will be added to
programs run on this QVM.
"""
            )

        self.noise_model = device.noise_model if device else None
        self.compiler = QVMCompiler(endpoint=compiler_endpoint, device=device) if device else None

        self.sync_endpoint = endpoint

        validate_noise_probabilities(gate_noise)
        validate_noise_probabilities(measurement_noise)
        self.gate_noise = gate_noise
        self.measurement_noise = measurement_noise

        if random_seed is None:
            self.random_seed = None
        elif isinstance(random_seed, int) and random_seed >= 0:
            self.random_seed = random_seed
        else:
            raise TypeError("random_seed should be None or a non-negative int")

        self._connection = ForestConnection(sync_endpoint=endpoint)
        self.session = self._connection.session  # backwards compatibility
        self.connect()

    def connect(self) -> None:
        try:
            version_dict = self.get_version_info()
            check_qvm_version(version_dict)
        except ConnectionError:
            raise QVMNotRunning(f"No QVM server running at {self._connection.sync_endpoint}")

    @_record_call
    def get_version_info(self) -> str:
        """
        Return version information for the QVM.

        :return: String with version information
        """
        return cast(str, self._connection._qvm_get_version_info())

    @_record_call
    def run(
        self,
        quil_program: Program,
        classical_addresses: Optional[Sequence[int]] = None,
        trials: int = 1,
    ) -> List[List[int]]:
        """
        Run a Quil program multiple times, accumulating the values deposited in
        a list of classical addresses.

        :param quil_program: A Quil program.
        :param classical_addresses: The classical memory to retrieve. Specified as a list of
            integers that index into a readout register named ``ro``. This function--and
            particularly this argument--are included for backwards compatibility and will
            be removed in the future.
        :param trials: Number of shots to collect.
        :return: A list of dictionaries of bits. Each dictionary corresponds to the values in
            `classical_addresses`.
        """
        if classical_addresses is None:
            caddresses: Mapping[str, Sequence[int]] = get_classical_addresses_from_program(
                quil_program
            )

        else:
            caddresses = {"ro": classical_addresses}

        buffers = self._connection._qvm_run(
            quil_program,
            caddresses,
            trials,
            self.measurement_noise,
            self.gate_noise,
            self.random_seed,
        )

        if len(buffers) == 0:
            return []
        if "ro" in buffers:
            return cast(List[List[int]], buffers["ro"].tolist())

        raise ValueError(
            "You are using QVMConnection.run with multiple readout registers not "
            "named `ro`. Please use the new `QuantumComputer` abstraction."
        )

    @_record_call
    def run_and_measure(
        self, quil_program: Program, qubits: Sequence[int], trials: int = 1
    ) -> List[List[int]]:
        """
        Run a Quil program once to determine the final wavefunction, and measure multiple times.

        :note: If the execution of ``quil_program`` is **non-deterministic**, i.e., if it includes
            measurements and/or noisy quantum gates, then the final wavefunction from which the
            returned bitstrings are sampled itself only represents a stochastically generated sample
            and the outcomes sampled from *different* ``run_and_measure`` calls *generally sample
            different bitstring distributions*.

        :param quil_program: A Quil program.
        :param qubits: A list of qubits.
        :param trials: Number of shots to collect.
        :return: A list of a list of bits.
        """
        # Developer note: This code is for backwards compatibility. It can't be replaced with
        # ForestConnection._run_and_measure because we've turned off the ability to set
        # `needs_compilation` (that usually indicates the user is doing something iffy like
        # using a noise model with this function)

        payload = self._run_and_measure_payload(quil_program, qubits, trials)
        assert self.sync_endpoint is not None
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return cast(List[List[int]], response.json())

    @_record_call
    def _run_and_measure_payload(
        self, quil_program: Program, qubits: Sequence[int], trials: int
    ) -> Dict[str, Any]:
        if not quil_program:
            raise ValueError(
                "You have attempted to run an empty program."
                " Please provide gates or measure instructions to your program."
            )

        if not isinstance(quil_program, Program):
            raise TypeError("quil_program must be a Quil program object")
        qubits = validate_qubit_list(qubits)
        if not isinstance(trials, int):
            raise TypeError("trials must be an integer")

        if self.noise_model is not None:
            assert self.compiler is not None
            compiled_program = self.compiler.quil_to_native_quil(quil_program)
            quil_program = apply_noise_model(compiled_program, self.noise_model)

        payload = {
            "type": TYPE_MULTISHOT_MEASURE,
            "qubits": list(qubits),
            "trials": trials,
            "compiled-quil": quil_program.out(),
        }

        self._maybe_add_noise_to_payload(payload)
        self._add_rng_seed_to_payload(payload)

        return payload

    @_record_call
    def wavefunction(self, quil_program: Program) -> Wavefunction:
        """
        Simulate a Quil program and get the wavefunction back.

        :note: If the execution of ``quil_program`` is **non-deterministic**, i.e., if it includes
            measurements and/or noisy quantum gates, then the final wavefunction from which the
            returned bitstrings are sampled itself only represents a stochastically generated sample
            and the wavefunctions returned by *different* ``wavefunction`` calls *will generally be
            different*.

        :param quil_program: A Quil program.
        :return: A Wavefunction object representing the state of the QVM.
        """
        # Developer note: This code is for backwards compatibility. It can't be replaced with
        # ForestConnection._wavefunction because we've turned off the ability to set
        # `needs_compilation` (that usually indicates the user is doing something iffy like
        # using a noise model with this function)

        payload = self._wavefunction_payload(quil_program)
        assert self.sync_endpoint is not None
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return Wavefunction.from_bit_packed_string(response.content)

    @_record_call
    def _wavefunction_payload(self, quil_program: Program) -> Dict[str, Any]:
        # Developer note: This code is for backwards compatibility. It can't be replaced with
        # _base_connection._wavefunction_payload because we've turned off the ability to set
        # `needs_compilation` (that usually indicates the user is doing something iffy like
        # using a noise model with this function)
        if not isinstance(quil_program, Program):
            raise TypeError("quil_program must be a Quil program object")

        payload = {"type": TYPE_WAVEFUNCTION, "compiled-quil": quil_program.out()}

        self._maybe_add_noise_to_payload(payload)
        self._add_rng_seed_to_payload(payload)

        return payload

    @_record_call
    def expectation(
        self, prep_prog: Program, operator_programs: Optional[Iterable[Program]] = None
    ) -> List[float]:
        """
        Calculate the expectation value of operators given a state prepared by
        prep_program.

        :note: If the execution of ``quil_program`` is **non-deterministic**, i.e., if it includes
            measurements and/or noisy quantum gates, then the final wavefunction from which the
            expectation values are computed itself only represents a stochastically generated
            sample. The expectations returned from *different* ``expectation`` calls *will then
            generally be different*.

        To measure the expectation of a PauliSum, you probably want to
        do something like this::

                progs, coefs = hamiltonian.get_programs()
                expect_coeffs = np.array(cxn.expectation(prep_program, operator_programs=progs))
                return np.real_if_close(np.dot(coefs, expect_coeffs))

        :param prep_prog: Quil program for state preparation.
        :param operator_programs: A list of Programs, each specifying an operator whose
            expectation to compute. Default is a list containing only the empty Program.
        :return: Expectation values of the operators.
        """
        # Developer note: This code is for backwards compatibility. It can't be replaced with
        # ForestConnection._expectation because we've turned off the ability to set
        # `needs_compilation` (that usually indicates the user is doing something iffy like
        # using a noise model with this function)

        if isinstance(operator_programs, Program):
            warnings.warn(
                "You have provided a Program rather than a list of Programs. The results from "
                "expectation will be line-wise expectation values of the operator_programs.",
                SyntaxWarning,
            )

        payload = self._expectation_payload(prep_prog, operator_programs)
        assert self.sync_endpoint is not None
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return cast(List[float], response.json())

    @_record_call
    def pauli_expectation(
        self, prep_prog: Program, pauli_terms: Union[Sequence[PauliTerm], PauliSum]
    ) -> Union[float, List[float]]:
        """
        Calculate the expectation value of Pauli operators given a state prepared by prep_program.

        If ``pauli_terms`` is a ``PauliSum`` then the returned value is a single ``float``,
        otherwise the returned value is a list of ``float``s, one for each ``PauliTerm`` in the
        list.

        :note: If the execution of ``quil_program`` is **non-deterministic**, i.e., if it includes
            measurements and/or noisy quantum gates, then the final wavefunction from which the
            expectation values are computed itself only represents a stochastically generated
            sample. The expectations returned from *different* ``expectation`` calls *will then
            generally be different*.

        :param Program prep_prog: Quil program for state preparation.
        :param Sequence[PauliTerm]|PauliSum pauli_terms: A list of PauliTerms or a PauliSum.
        :return: If ``pauli_terms`` is a PauliSum return its expectation value. Otherwise return
          a list of expectation values.
        :rtype: float|List[float]
        """

        is_pauli_sum = False
        if isinstance(pauli_terms, PauliSum):
            progs, coeffs = pauli_terms.get_programs()
            is_pauli_sum = True
        else:
            coeffs = [pt.coefficient for pt in pauli_terms]
            progs = [pt.program for pt in pauli_terms]

        bare_results = self.expectation(prep_prog, progs)
        results = [c * r for c, r in zip(coeffs, bare_results)]
        if is_pauli_sum:
            return sum(results)
        return results

    def _expectation_payload(
        self, prep_prog: Program, operator_programs: Optional[Iterable[Program]]
    ) -> Dict[str, Any]:
        if operator_programs is None:
            operator_programs = [Program()]

        if not isinstance(prep_prog, Program):
            raise TypeError("prep_prog variable must be a Quil program object")

        payload = {
            "type": TYPE_EXPECTATION,
            "state-preparation": prep_prog.out(),
            "operators": [x.out() for x in operator_programs],
        }

        self._add_rng_seed_to_payload(payload)

        return payload

    def _maybe_add_noise_to_payload(self, payload: Dict[str, Any]) -> None:
        """
        Set the gate noise and measurement noise of a payload.
        """
        if self.measurement_noise is not None:
            payload["measurement-noise"] = self.measurement_noise
        if self.gate_noise is not None:
            payload["gate-noise"] = self.gate_noise

    def _add_rng_seed_to_payload(self, payload: Dict[str, Any]) -> None:
        """
        Add a random seed to the payload.
        """
        if self.random_seed is not None:
            payload["rng-seed"] = self.random_seed


class QVM(QAM):
    @_record_call
    def __init__(
        self,
        connection: ForestConnection,
        noise_model: Optional[NoiseModel] = None,
        gate_noise: Optional[List[float]] = None,
        measurement_noise: Optional[List[float]] = None,
        random_seed: Optional[int] = None,
        requires_executable: bool = False,
    ) -> None:
        """
        A virtual machine that classically emulates the execution of Quil programs.

        :param connection: A connection to the Forest web API.
        :param noise_model: A noise model that describes noise to apply when emulating a program's
            execution.
        :param gate_noise: A list of three numbers [Px, Py, Pz] indicating the probability of an X,
           Y, or Z gate getting applied to each qubit after a gate application or reset. The
           default value of None indicates no noise.
        :param measurement_noise: A list of three numbers [Px, Py, Pz] indicating the probability
            of an X, Y, or Z gate getting applied before a measurement. The default value of
            None indicates no noise.
        :param random_seed: A seed for the QVM's random number generators. Either None (for an
            automatically generated seed) or a non-negative integer.
        :param requires_executable: Whether this QVM will refuse to run a :py:class:`Program` and
            only accept the result of :py:func:`compiler.native_quil_to_executable`. Setting this
            to True better emulates the behavior of a QPU.
        """
        super().__init__()

        if (noise_model is not None) and (gate_noise is not None or measurement_noise is not None):
            raise ValueError(
                """
You have attempted to supply the QVM with both a Kraus noise model
(by supplying a `noise_model` argument), as well as either `gate_noise`
or `measurement_noise`. At this time, only one may be supplied.

To read more about supplying noise to the QVM, see
http://pyquil.readthedocs.io/en/latest/noise_models.html#support-for-noisy-gates-on-the-rigetti-qvm.
"""
            )

        self.noise_model = noise_model
        self.connection = connection

        validate_noise_probabilities(gate_noise)
        validate_noise_probabilities(measurement_noise)
        self.gate_noise = gate_noise
        self.measurement_noise = measurement_noise

        if random_seed is None:
            self.random_seed = None
        elif isinstance(random_seed, int) and random_seed >= 0:
            self.random_seed = random_seed
        else:
            raise TypeError("random_seed should be None or a non-negative int")

        self.requires_executable = requires_executable
        self.connect()

    def connect(self) -> None:
        try:
            version_dict = self.get_version_info()
            check_qvm_version(version_dict)
        except ConnectionError:
            raise QVMNotRunning(f"No QVM server running at {self.connection.sync_endpoint}")

    @_record_call
    def get_version_info(self) -> str:
        """
        Return version information for the QVM.

        :return: String with version information
        """
        return cast(str, self.connection._qvm_get_version_info())

    @_record_call
    def load(self, executable: Union[Program, PyQuilExecutableResponse]) -> "QVM":
        """
        Initialize a QAM and load a program to be executed with a call to :py:func:`run`.

        If ``QVM.requires_executable`` is set to ``True``, this function will only load
        :py:class:`PyQuilExecutableResponse` executables. This more closely follows the behavior
        of :py:class:`QPU`. However, the quantum simulator doesn't *actually* need a compiled
        binary executable, so if this flag is set to ``False`` we also accept :py:class:`Program`
        objects.

        :param executable: An executable. See the above note for acceptable types.
        """
        if self.requires_executable:
            if isinstance(executable, PyQuilExecutableResponse):
                executable = _extract_program_from_pyquil_executable_response(executable)
            else:
                raise TypeError(
                    "`executable` argument must be a `PyQuilExecutableResponse`. Make "
                    "sure you have explicitly compiled your program via `qc.compile` "
                    "or `qc.compiler.native_quil_to_executable(...)` for more "
                    "fine-grained control. This explicit step is required for running "
                    "on a QPU."
                )
        else:
            if isinstance(executable, PyQuilExecutableResponse):
                executable = _extract_program_from_pyquil_executable_response(executable)
            elif isinstance(executable, Program):
                pass
            else:
                raise TypeError(
                    "`executable` argument must be a `PyQuilExecutableResponse` or a "
                    "`Program`. You provided {}".format(type(executable))
                )

        return cast("QVM", super().load(executable))

    @_record_call
    def run(self) -> "QVM":
        """
        Run a Quil program on the QVM multiple times and return the values stored in the
        classical registers designated by the classical_addresses parameter.

        :return: An array of bitstrings of shape ``(trials, len(classical_addresses))``
        """

        super().run()

        if not isinstance(self._executable, Program):
            # This should really never happen
            # unless a user monkeys with `self.status` and `self._executable`.
            raise ValueError("Please `load` an appropriate executable.")

        quil_program = self._executable
        trials = quil_program.num_shots
        classical_addresses = get_classical_addresses_from_program(quil_program)

        if self.noise_model is not None:
            quil_program = apply_noise_model(quil_program, self.noise_model)

        quil_program = self.augment_program_with_memory_values(quil_program)

        self._memory_results = self.connection._qvm_run(
            quil_program=quil_program,
            classical_addresses=classical_addresses,
            trials=trials,
            measurement_noise=self.measurement_noise,
            gate_noise=self.gate_noise,
            random_seed=self.random_seed,
        )

        if "ro" not in self._memory_results or len(self._memory_results["ro"]) == 0:
            self._memory_results["ro"] = np.zeros((trials, 0), dtype=np.int64)

        return self

    def augment_program_with_memory_values(self, quil_program: Program) -> Program:
        p = Program()

        for k, v in self._variables_shim.items():
            p += MOVE(MemoryReference(name=k.name, offset=k.index), v)

        p += quil_program

        return percolate_declares(p)

    @_record_call
    def reset(self) -> None:
        """
        Reset the state of the underlying QAM, and the QVM connection information.
        """
        super().reset()
        forest_connection = ForestConnection()
        self.connection = forest_connection



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import sys
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast

import networkx as nx
import numpy as np

from pyquil.quilatom import Parameter, unpack_qubit
from pyquil.quilbase import Gate

if sys.version_info < (3, 7):
    from pyquil.external.dataclasses import dataclass
else:
    from dataclasses import dataclass

DEFAULT_QUBIT_TYPE = "Xhalves"
DEFAULT_EDGE_TYPE = "CZ"
THETA = Parameter("theta")
"Used as the symbolic parameter in RZ, CPHASE gates."


@dataclass
class MeasureInfo:
    operator: Optional[str] = None
    qubit: Optional[Union[int, str]] = None
    target: Optional[Union[int, str]] = None
    duration: Optional[float] = None
    fidelity: Optional[float] = None


@dataclass
class GateInfo:
    operator: Optional[str] = None
    parameters: Optional[Sequence[Union[str, float]]] = None
    arguments: Optional[Sequence[Union[str, float]]] = None
    duration: Optional[float] = None
    fidelity: Optional[float] = None


@dataclass
class Qubit:
    id: int
    type: Optional[str] = None
    dead: Optional[bool] = None
    gates: Optional[Sequence[Union[GateInfo, MeasureInfo]]] = None


@dataclass
class Edge:
    targets: Tuple[int, ...]
    type: Optional[Union[List[str], str]] = None
    dead: Optional[bool] = None
    gates: Optional[Sequence[GateInfo]] = None


@dataclass
class ISA:
    """
    Basic Instruction Set Architecture specification.

    :ivar qubits: The qubits associated with the ISA.
    :ivar edges: The multi-qubit gates.
    """

    qubits: Sequence[Qubit]
    edges: Sequence[Edge]

    def to_dict(self) -> Dict[str, Any]:
        """
        Create a JSON-serializable representation of the ISA.

        The dictionary representation is of the form::

            {
                "1Q": {
                    "0": {
                        "type": "Xhalves"
                    },
                    "1": {
                        "type": "Xhalves",
                        "dead": True
                    },
                    ...
                },
                "2Q": {
                    "1-4": {
                        "type": "CZ"
                    },
                    "1-5": {
                        "type": "CZ"
                    },
                    ...
                },
                ...
            }

        :return: A dictionary representation of self.
        """

        def _maybe_configure(o: Union[Qubit, Edge], t: Union[str, List[str]]) -> Dict[str, Any]:
            """
            Exclude default values from generated dictionary.

            :param o: The object to serialize
            :param t: The default value for ``o.type``.
            :return: d
            """
            d: Dict[str, Any] = {}

            if o.gates is None or len(o.gates) == 0:
                inferred_type = o.type if (o.type is not None and o.type != t) else t
                inferred_gates = convert_gate_type_to_gate_information(inferred_type)
            else:
                inferred_gates = cast(List[Union[GateInfo, MeasureInfo]], o.gates)

            d["gates"] = [
                {
                    "operator": i.operator,
                    "parameters": i.parameters,
                    "arguments": i.arguments,
                    "fidelity": i.fidelity,
                    "duration": i.duration,
                }
                if isinstance(i, GateInfo)
                else {
                    "operator": "MEASURE",
                    "qubit": i.qubit,
                    "target": i.target,
                    "duration": i.duration,
                    "fidelity": i.fidelity,
                }
                for i in inferred_gates
            ]
            if o.dead:
                d["dead"] = o.dead
            return d

        return {
            "1Q": {"{}".format(q.id): _maybe_configure(q, DEFAULT_QUBIT_TYPE) for q in self.qubits},
            "2Q": {
                "{}-{}".format(*edge.targets): _maybe_configure(edge, DEFAULT_EDGE_TYPE)
                for edge in self.edges
            },
        }

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ISA":
        """
        Re-create the ISA from a dictionary representation.

        :param d: The dictionary representation.
        :return: The restored ISA.
        """
        return ISA(
            qubits=sorted(
                [
                    Qubit(
                        id=int(qid),
                        type=q.get("type", DEFAULT_QUBIT_TYPE),
                        dead=q.get("dead", False),
                    )
                    for qid, q in d["1Q"].items()
                ],
                key=lambda qubit: qubit.id,
            ),
            edges=sorted(
                [
                    Edge(
                        targets=tuple(int(q) for q in eid.split("-")),
                        type=e.get("type", DEFAULT_EDGE_TYPE),
                        dead=e.get("dead", False),
                    )
                    for eid, e in d["2Q"].items()
                ],
                key=lambda edge: edge.targets,
            ),
        )


def convert_gate_type_to_gate_information(
    gate_type: Union[str, List[str]]
) -> List[Union[GateInfo, MeasureInfo]]:
    if isinstance(gate_type, str):
        gate_type = [gate_type]

    gate_information: List[Union[GateInfo, MeasureInfo]] = []

    for type_keyword in gate_type:
        if type_keyword == "Xhalves":
            gate_information.extend(
                [
                    GateInfo("I", [], ["_"]),
                    GateInfo("RX", [np.pi / 2], ["_"]),
                    GateInfo("RX", [-np.pi / 2], ["_"]),
                    GateInfo("RX", [np.pi], ["_"]),
                    GateInfo("RX", [-np.pi], ["_"]),
                    GateInfo("RZ", ["theta"], ["_"]),
                    MeasureInfo(operator="MEASURE", qubit="_", target="_"),
                    MeasureInfo(operator="MEASURE", qubit="_", target=None),
                ]
            )
        elif type_keyword == "WILDCARD":
            gate_information.extend([GateInfo("_", "_", ["_"]), GateInfo("_", "_", ["_", "_"])])
        elif type_keyword == "CZ":
            gate_information.extend([GateInfo("CZ", [], ["_", "_"])])
        elif type_keyword == "ISWAP":
            gate_information.extend([GateInfo("ISWAP", [], ["_", "_"])])
        elif type_keyword == "CPHASE":
            gate_information.extend([GateInfo("CPHASE", ["theta"], ["_", "_"])])
        elif type_keyword == "XY":
            gate_information.extend([GateInfo("XY", ["theta"], ["_", "_"])])
        else:
            raise ValueError("Unknown edge type: {}".format(type_keyword))

    return gate_information


def gates_in_isa(isa: ISA) -> List[Gate]:
    """
    Generate the full gateset associated with an ISA.

    :param isa: The instruction set architecture for a QPU.
    :return: A sequence of Gate objects encapsulating all gates compatible with the ISA.
    """
    gates = []
    for q in isa.qubits:
        if q.dead:
            # TODO: dead qubits may in the future lead to some implicit re-indexing
            continue
        if q.type == "Xhalves":
            gates.extend(
                [
                    Gate("I", [], [unpack_qubit(q.id)]),
                    Gate("RX", [np.pi / 2], [unpack_qubit(q.id)]),
                    Gate("RX", [-np.pi / 2], [unpack_qubit(q.id)]),
                    Gate("RX", [np.pi], [unpack_qubit(q.id)]),
                    Gate("RX", [-np.pi], [unpack_qubit(q.id)]),
                    Gate("RZ", [THETA], [unpack_qubit(q.id)]),
                ]
            )
        elif q.type == "WILDCARD":
            gates.extend([Gate("_", "_", [unpack_qubit(q.id)])])
        else:  # pragma no coverage
            raise ValueError("Unknown qubit type: {}".format(q.type))

    for e in isa.edges:
        if e.dead:
            continue
        targets = [unpack_qubit(t) for t in e.targets]
        assert e.type is not None
        edge_type = e.type if isinstance(e.type, list) else [e.type]
        if "CZ" in edge_type:
            gates.append(Gate("CZ", [], targets))
            gates.append(Gate("CZ", [], targets[::-1]))
            continue
        if "ISWAP" in edge_type:
            gates.append(Gate("ISWAP", [], targets))
            gates.append(Gate("ISWAP", [], targets[::-1]))
            continue
        if "CPHASE" in edge_type:
            gates.append(Gate("CPHASE", [THETA], targets))
            gates.append(Gate("CPHASE", [THETA], targets[::-1]))
            continue
        if "XY" in edge_type:
            gates.append(Gate("XY", [THETA], targets))
            gates.append(Gate("XY", [THETA], targets[::-1]))
            continue
        assert e.type is not None
        if "WILDCARD" in e.type:
            gates.append(Gate("_", "_", targets))
            gates.append(Gate("_", "_", targets[::-1]))
            continue

        raise ValueError("Unknown edge type: {}".format(e.type))
    return gates


def isa_from_graph(
    graph: nx.Graph, oneq_type: str = "Xhalves", twoq_type: Optional[Union[str, List[str]]] = None
) -> ISA:
    """
    Generate an ISA object from a NetworkX graph.

    :param graph: The graph
    :param oneq_type: The type of 1-qubit gate. Currently 'Xhalves'
    :param twoq_type: The type of 2-qubit gate. One or more of 'CZ', 'CPHASE', 'ISWAP', 'XY'.
                      The default, None, is a synonym for ["CZ", "XY"].
    """
    all_qubits = list(range(max(graph.nodes) + 1))
    qubits = [Qubit(i, type=oneq_type, dead=i not in graph.nodes) for i in all_qubits]
    edges = [
        Edge(
            tuple(sorted((a, b))), type=["CZ", "XY"] if twoq_type is None else twoq_type, dead=False
        )
        for a, b in graph.edges
    ]
    return ISA(qubits, edges)


def isa_to_graph(isa: ISA) -> nx.Graph:
    """
    Construct a NetworkX qubit topology from an ISA object.

    This discards information about supported gates.

    :param isa: The ISA.
    """
    return nx.from_edgelist(e.targets for e in isa.edges if not e.dead)



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for automatically generating error reports helpful for diagnosing pyQuil errors.

IMPORTANT NOTE: THIS MODULE USES GLOBAL STATE AND IS NOT ESPECIALLY THREAD-SAFE.
                If your threaded code is experiencing pyQuil errors, you'll have to track
                your own state and not use this convenient decorator.
"""
import inspect
import json
import logging
import os
import sys
from datetime import datetime, date
from functools import wraps
from typing import List, Dict, Any, Callable, Optional

from pyquil.quil import Program
from pyquil.version import __version__

if sys.version_info < (3, 7):
    from pyquil.external.dataclasses import dataclass, is_dataclass, asdict
else:
    from dataclasses import dataclass, is_dataclass, asdict

_log = logging.getLogger(__name__)


@dataclass
class ErrorReport:
    """
    Dump of the current state of a pyQuil program.
    """

    stack_trace: List["StacktraceFrame"]
    timestamp: date
    call_log: Dict[str, "CallLogValue"]
    exception: Exception  # noqa: E701
    system_info: Dict[str, str]


@dataclass
class StacktraceFrame:
    """
    Expanded frame in a stacktrace, suitable for JSON export.
    """

    name: str
    filename: str
    line_number: int
    locals: Dict[str, str]


@dataclass(eq=True, frozen=True)
class CallLogKey:
    """
    Entry in the call log list, suitable for JSON export.
    """

    name: str
    args: List[str]
    kwargs: Dict[str, Any]

    def __hash__(self) -> int:
        finger_print = (
            (self.name,) + tuple(self.args) + tuple(sorted(self.kwargs.items(), key=lambda i: i[0]))
        )
        return hash(finger_print)

    def __repr__(self) -> str:
        ret = self.name + "("
        for item in self.args:
            ret += repr(item) + ", "
        for k, v in self.kwargs.items():
            ret += k + "=" + repr(v) + ", "
        ret += ")"

        return ret


@dataclass
class CallLogValue:
    """
    Entry in the call log list, suitable for JSON export.
    """

    timestamp_in: date
    timestamp_out: Optional[date]
    return_value: Optional[str]


def json_serialization_helper(o: object) -> Any:
    if is_dataclass(o):
        return asdict(o)
    elif isinstance(o, datetime):
        return o.isoformat()
    elif isinstance(o, Exception):
        return repr(o)
    else:
        raise TypeError("unable to serialize object {}".format(o))


def generate_system_info() -> Dict[str, str]:
    system_info = {"python_version": sys.version, "pyquil_version": __version__}

    return system_info


def serialize_object_for_logging(o: object) -> str:
    if isinstance(o, Program):
        return str(o)
    else:
        return repr(o)


def flatten_log(log: Dict[CallLogKey, CallLogValue]) -> Dict[str, CallLogValue]:
    return {repr(k): v for k, v in log.items()}


class ErrorContext(object):
    """
    Tracks information relevant to error reporting.
    """

    log: Dict[CallLogKey, CallLogValue] = {}
    filename = "pyquil_error.log"

    def generate_report(self, exception: Exception, trace: List[inspect.FrameInfo]) -> ErrorReport:
        """
        Handle an error generated in a routine decorated with the pyQuil error handler.

        :param exception: Exception object that generated this error.
        :param trace: inspect.trace object from the frame that caught the error.
        :return: ErrorReport object
        """
        stack_trace = [
            StacktraceFrame(
                name=item.function,
                filename=item.filename,
                line_number=item.lineno,
                locals={
                    k: serialize_object_for_logging(v) for (k, v) in item.frame.f_locals.items()
                },
            )
            for item in trace
        ]

        system_info = generate_system_info()

        report = ErrorReport(
            stack_trace=stack_trace,
            timestamp=datetime.utcnow(),
            exception=exception,
            system_info=system_info,
            call_log=flatten_log(self.log),
        )

        return report

    def dump_error(self, exception: Exception, trace: List[inspect.FrameInfo]) -> None:
        warn_msg = """
>>> PYQUIL_PROTECT <<<
An uncaught exception was raised in a function wrapped in pyquil_protect.  We are writing out a
log file to "{}".

Along with a description of what you were doing when the error occurred, send this file to
Rigetti Computing support by email at support@rigetti.com for assistance.
>>> PYQUIL_PROTECT <<<
""".format(
            os.path.abspath(self.filename)
        )

        _log.warning(warn_msg)

        report = self.generate_report(exception, trace)

        # overwrite any existing log file
        fh = open(self.filename, "w")
        fh.write(json.dumps(report, default=json_serialization_helper))
        fh.close()


global_error_context: Optional[ErrorContext] = None


def pyquil_protect(
    func: Callable[..., Any], log_filename: str = "pyquil_error.log"
) -> Callable[..., Any]:
    """
    A decorator that sets up an error context, captures errors, and tears down the context.
    """

    def pyquil_protect_wrapper(*args: Any, **kwargs: Any) -> Any:
        global global_error_context

        old_error_context = global_error_context
        global_error_context = ErrorContext()
        global_error_context.filename = log_filename

        try:
            val = func(*args, **kwargs)
            global_error_context = old_error_context
            return val
        except Exception as e:
            assert global_error_context is not None
            global_error_context.dump_error(e, inspect.trace())
            global_error_context = old_error_context
            raise

    return pyquil_protect_wrapper


def _record_call(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    A decorator that logs a call into the global error context.

    This is probably for internal use only.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        global global_error_context

        # log a call as about to take place
        if global_error_context is not None:
            key = CallLogKey(
                name=func.__name__,
                args=[serialize_object_for_logging(arg) for arg in args],
                kwargs={k: serialize_object_for_logging(v) for k, v in kwargs.items()},
            )

            pre_entry = CallLogValue(
                timestamp_in=datetime.utcnow(), timestamp_out=None, return_value=None
            )
            global_error_context.log[key] = pre_entry

        val = func(*args, **kwargs)

        # poke the return value of that call in
        if global_error_context is not None:
            post_entry = CallLogValue(
                timestamp_in=pre_entry.timestamp_in,
                timestamp_out=datetime.utcnow(),
                return_value=serialize_object_for_logging(val),
            )
            global_error_context.log[key] = post_entry

        return val

    return wrapper



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module containing the Wavefunction object and methods for working with wavefunctions.
"""
import itertools
import struct
import warnings
from typing import Dict, Iterator, List, Optional, Sequence, cast

import numpy as np

OCTETS_PER_DOUBLE_FLOAT = 8
OCTETS_PER_COMPLEX_DOUBLE = 2 * OCTETS_PER_DOUBLE_FLOAT


class Wavefunction(object):
    """
    Encapsulate a wavefunction representing a quantum state
    as returned by :py:class:`~pyquil.api.WavefunctionSimulator`.

    .. note::

        The elements of the wavefunction are ordered by bitstring. E.g., for two qubits the order
        is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
        i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
        See also :ref:`the related documentation section in the WavefunctionSimulator Overview
        <basis_ordering>`.
    """

    def __init__(self, amplitude_vector: np.ndarray):
        """
        Initializes a wavefunction

        :param amplitude_vector: A numpy array of complex amplitudes
        """
        if len(amplitude_vector) == 0 or len(amplitude_vector) & (len(amplitude_vector) - 1) != 0:
            raise TypeError("Amplitude vector must have a length that is a power of two")

        self.amplitudes = np.asarray(amplitude_vector)
        sumprob = np.sum(self.probabilities())
        if not np.isclose(sumprob, 1.0):
            raise ValueError(
                "The wavefunction is not normalized. "
                "The probabilities sum to {} instead of 1".format(sumprob)
            )

    @staticmethod
    def ground(qubit_num: int) -> "Wavefunction":
        warnings.warn("ground() has been deprecated in favor of zeros()", stacklevel=2)
        return Wavefunction.zeros(qubit_num)

    @staticmethod
    def zeros(qubit_num: int) -> "Wavefunction":
        """
        Constructs the groundstate wavefunction for a given number of qubits.

        :param qubit_num:
        :return: A Wavefunction in the ground state
        """
        amplitude_vector = np.zeros(2 ** qubit_num)
        amplitude_vector[0] = 1.0
        return Wavefunction(amplitude_vector)

    @staticmethod
    def from_bit_packed_string(coef_string: bytes) -> "Wavefunction":
        """
        From a bit packed string, unpacks to get the wavefunction
        :param coef_string:
        """
        num_octets = len(coef_string)

        # Parse the wavefunction
        wf = np.zeros(num_octets // OCTETS_PER_COMPLEX_DOUBLE, dtype=np.cfloat)
        for i, p in enumerate(range(0, num_octets, OCTETS_PER_COMPLEX_DOUBLE)):
            re_be = coef_string[p : p + OCTETS_PER_DOUBLE_FLOAT]
            im_be = coef_string[p + OCTETS_PER_DOUBLE_FLOAT : p + OCTETS_PER_COMPLEX_DOUBLE]
            re = struct.unpack(">d", re_be)[0]
            im = struct.unpack(">d", im_be)[0]
            wf[i] = complex(re, im)

        return Wavefunction(wf)

    def __len__(self) -> int:
        return len(self.amplitudes).bit_length() - 1

    def __iter__(self) -> Iterator[complex]:
        return cast(Iterator[complex], self.amplitudes.__iter__())

    def __getitem__(self, index: int) -> complex:
        return cast(complex, self.amplitudes[index])

    def __setitem__(self, key: int, value: complex) -> None:
        self.amplitudes[key] = value

    def __str__(self) -> str:
        return self.pretty_print(decimal_digits=10)

    def probabilities(self) -> np.ndarray:
        """Returns an array of probabilities in lexicographical order"""
        return np.abs(self.amplitudes) ** 2

    def get_outcome_probs(self) -> Dict[str, float]:
        """
        Parses a wavefunction (array of complex amplitudes) and returns a dictionary of
        outcomes and associated probabilities.

        :return: A dict with outcomes as keys and probabilities as values.
        :rtype: dict
        """
        outcome_dict = {}
        qubit_num = len(self)
        for index, amplitude in enumerate(self.amplitudes):
            outcome = get_bitstring_from_index(index, qubit_num)
            outcome_dict[outcome] = abs(amplitude) ** 2
        return outcome_dict

    def pretty_print_probabilities(self, decimal_digits: int = 2) -> Dict[str, float]:
        """
        TODO: This doesn't seem like it is named correctly...

        Prints outcome probabilities, ignoring all outcomes with approximately zero probabilities
        (up to a certain number of decimal digits) and rounding the probabilities to decimal_digits.

        :param int decimal_digits: The number of digits to truncate to.
        :return: A dict with outcomes as keys and probabilities as values.
        """
        outcome_dict = {}
        qubit_num = len(self)
        for index, amplitude in enumerate(self.amplitudes):
            outcome = get_bitstring_from_index(index, qubit_num)
            prob = round(abs(amplitude) ** 2, decimal_digits)
            if prob != 0.0:
                outcome_dict[outcome] = prob
        return outcome_dict

    def pretty_print(self, decimal_digits: int = 2) -> str:
        """
        Returns a string repr of the wavefunction, ignoring all outcomes with approximately zero
        amplitude (up to a certain number of decimal digits) and rounding the amplitudes to
        decimal_digits.

        :param int decimal_digits: The number of digits to truncate to.
        :return: A string representation of the wavefunction.
        """
        outcome_dict = {}
        qubit_num = len(self)
        pp_string = ""
        for index, amplitude in enumerate(self.amplitudes):
            outcome = get_bitstring_from_index(index, qubit_num)
            amplitude = (
                round(amplitude.real, decimal_digits) + round(amplitude.imag, decimal_digits) * 1.0j
            )
            if amplitude != 0.0:
                outcome_dict[outcome] = amplitude
                pp_string += str(amplitude) + "|{}> + ".format(outcome)
        if len(pp_string) >= 3:
            pp_string = pp_string[:-3]  # remove the dangling + if it is there
        return pp_string

    def plot(self, qubit_subset: Optional[Sequence[int]] = None) -> None:
        """
        TODO: calling this will error because of matplotlib

        Plots a bar chart with bitstring on the x axis and probability on the y axis.

        :param qubit_subset: Optional parameter used for plotting a subset of the Hilbert space.
        """
        import matplotlib.pyplot as plt

        prob_dict = self.get_outcome_probs()
        if qubit_subset:
            sub_dict = {}
            qubit_num = len(self)
            for i in qubit_subset:
                if i > (2 ** qubit_num - 1):
                    raise IndexError("Index {} too large for {} qubits.".format(i, qubit_num))
                else:
                    sub_dict[get_bitstring_from_index(i, qubit_num)] = prob_dict[
                        get_bitstring_from_index(i, qubit_num)
                    ]
            prob_dict = sub_dict
        plt.bar(range(len(prob_dict)), prob_dict.values(), align="center", color="#6CAFB7")
        plt.xticks(range(len(prob_dict)), prob_dict.keys())
        plt.show()

    def sample_bitstrings(self, n_samples: int) -> np.ndarray:
        """
        Sample bitstrings from the distribution defined by the wavefunction.

        :param n_samples: The number of bitstrings to sample
        :return: An array of shape (n_samples, n_qubits)
        """
        possible_bitstrings = np.array(list(itertools.product((0, 1), repeat=len(self))))
        inds = np.random.choice(2 ** len(self), n_samples, p=self.probabilities())
        bitstrings = possible_bitstrings[inds, :]
        return bitstrings


def get_bitstring_from_index(index: int, qubit_num: int) -> str:
    """
    Returns the bitstring in lexical order that corresponds to the given index in 0 to 2^(qubit_num)
    :param int index:
    :param int qubit_num:
    :return: the bitstring
    :rtype: str
    """
    if index > (2 ** qubit_num - 1):
        raise IndexError("Index {} too large for {} qubits.".format(index, qubit_num))
    return bin(index)[2:].rjust(qubit_num, "0")


def _octet_bits(o: int) -> List[int]:
    """
    Get the bits of an octet.

    :param o: The octets.
    :return: The bits as a list in LSB-to-MSB order.
    """
    if not isinstance(o, int):
        raise TypeError("o should be an int")
    if not (0 <= o <= 255):
        raise ValueError("o should be between 0 and 255 inclusive")
    bits = [0] * 8
    for i in range(8):
        if 1 == o & 1:
            bits[i] = 1
        o = o >> 1
    return bits
